<?php

declare(strict_types=1);

namespace Drupal\flysystem_local\Plugin\FlysystemAdapter;

use Drupal\Core\Form\FormStateInterface;
use Drupal\flysystem\FlysystemSubAdapterInterface;
use Drupal\flysystem\Plugin\FlysystemAdapterPluginBase;
use Drupal\flysystem\Plugin\FlysystemAdapterPluginTrait;
use League\Flysystem\Local\LocalFilesystemAdapter;

/**
 * Plugin implementation of the flysystem_adapter.
 *
 * @FlysystemAdapter(
 *   id = "local_adapter",
 *   label = @Translation("Local Adapter"),
 *   description = @Translation("Flysystem Local Adapter.")
 * )
 */
final class LocalAdapter extends FlysystemAdapterPluginBase implements FlysystemSubAdapterInterface {

  use FlysystemAdapterPluginTrait;

  /**
   * The flysystem_subadapter ID.
   */
  protected string $id;

  /**
   * The flysystem_subadapter label.
   */
  protected string $label;

  /**
   * The flysystem_subadapter description.
   */
  protected string $description;

  /**
   * Default permissions array.
   */
  protected array $defaultPermissions = [
    'file' => [
      'public' => '0664',
      'private' => '0664',
    ],
    'dir' => [
      'public' => '0775',
      'private' => '0775',
    ],
  ];

  /**
   * The League\Flysystem\FilesystemAdapter class for this adapter.
   */
  protected static string $flysystemAdapterClass = 'League\Flysystem\Local\LocalFilesystemAdapter';

  /**
   * The fully qualified name of this class.
   */
  protected string $pluginAdapterClass = __CLASS__;

  /**
   * {@inheritdoc}
   */
  public function getDescription() {
    return !empty($this->description) ? $this->description : '';
  }

  /**
   * {@inheritdoc}
   */
  public function buildSubConfigurationForm(array &$form, FormStateInterface $form_state, $flysystem_adapter_config) {
    $defaultPermissions = $this->getDefaultPermissions();
    $form['root_directory'] = [
      '#type' => 'textfield',
      '#title' => $this->t('File Storage Location'),
      '#maxlength' => 255,
      '#description' => $this->t('File Storage directory, relative to Drupal root.'),
      // @todo determine how to retrieve saved values once form is saved.
      '#default_value' => $flysystem_adapter_config->getSubAdapterConfigItem('root_directory'),
      '#required' => TRUE,
    ];
    $form['permissions']['file']['public'] = [
      '#type' => 'number',
      '#title' => $this->t('Public file permissions'),
      '#description' => $this->t('Default permissions for public files.'),
      // @todo determine how to retrieve saved values once form is saved.
      '#default_value' => $flysystem_adapter_config->getSubAdapterConfigItem(['permissions', 'file', 'public']) ?? $defaultPermissions['file']['public'],
    ];
    $form['permissions']['file']['private'] = [
      '#type' => 'number',
      '#title' => $this->t('Private file permissions'),
      '#description' => $this->t('Default permissions for private files.'),
      // @todo determine how to retrieve saved values once form is saved.
      '#default_value' => $flysystem_adapter_config->getSubAdapterConfigItem(['permissions', 'file', 'private']) ?? $defaultPermissions['file']['private'],
    ];
    $form['permissions']['dir']['public'] = [
      '#type' => 'number',
      '#title' => $this->t('Public directory permissions'),
      '#description' => $this->t('Default permissions for public directories.'),
      // @todo determine how to retrieve saved values once form is saved.
      '#default_value' => $flysystem_adapter_config->getSubAdapterConfigItem(['permissions', 'dir', 'public']) ?? $defaultPermissions['dir']['public'],
    ];
    $form['permissions']['dir']['private'] = [
      '#type' => 'number',
      '#title' => $this->t('Private directory permissions'),
      '#description' => $this->t('Default permissions for private directories.'),
      // @todo determine how to retrieve saved values once form is saved.
      '#default_value' => $flysystem_adapter_config->getSubAdapterConfigItem(['permissions', 'dir', 'private']) ?? $defaultPermissions['dir']['private'],
    ];
    $form['write_flags'] = [
      '#type' => 'hidden',
      '#title' => $this->t('Write flags'),
      // @todo convert to proper translated entity.
      '#description' => $this->t('Lock types for file resources, reference <a href="https://www.php.net/manual/en/function.flock.php#:~:text=operation-,operation,-is%20one%20of">PHP file locking operations</a> for more information.'),
      // @todo determine how to retrieve saved values once form is saved.
      '#default_value' => (strlen($flysystem_adapter_config->getSubAdapterConfigItem('write_flag')) == 0) ? LOCK_EX : NULL,
    ];
    $form['link_handling'] = [
      '#type' => 'hidden',
      '#title' => $this->t('Link handler flags'),
      // @todo convert to proper translated entity.
      '#description' => $this->t('How to handle links, reference <a href="https://flysystem.thephpleague.com/docs/adapter/local/">Local File Adapter, advanced usage</a> for more information.'),
      // @todo determine how to retrieve saved values once form is saved.
      '#default_value' => (strlen($flysystem_adapter_config->getSubAdapterConfigItem('link_handling')) == 0) ? LocalFilesystemAdapter::DISALLOW_LINKS : NULL,
    ];

    $form['flysystem_adapter_class'] = [
      '#type' => 'hidden',
      '#title' => $this->t('Flysystem Adapter Class'),
      '#default_value' => self::getFlysystemAdapterClass(),
    ];

    $form['plugin_adapter_class'] = [
      '#type' => 'hidden',
      '#title' => $this->t('Drupal Flysystem Plugin Adapter Class'),
      '#default_value' => __CLASS__ . "Utility",
    ];

    return $form;
  }

  /**
   * Return Flysystem Adapter class name.
   *
   * @return string
   *   Fully qualified class name for \League\Flysystem\FilesystemAdapter
   */
  public static function getFlysystemAdapterClass() {
    return self::$flysystemAdapterClass;
  }

  /**
   * Returns a configured Flysystem FilesystemAdapter.
   */
  public static function configureFlysystemAdapter(array $configure_entity) {

  }

}
