<?php

namespace Drupal\flysystem\Flysystem\Adapter;

use League\Flysystem\FilesystemAdapter;
use League\Flysystem\Config;
use League\Flysystem\FileAttributes;
use League\Flysystem\InvalidVisibilityProvided;
use League\Flysystem\UnableToDeleteDirectory;
use League\Flysystem\UnableToDeleteFile;
use League\Flysystem\UnableToReadFile;
use League\Flysystem\UnableToRetrieveMetadata;
use League\Flysystem\UnableToCheckExistence;
use League\Flysystem\UnableToCopyFile;
use League\Flysystem\UnableToCreateDirectory;
use League\Flysystem\UnableToListContents;
use League\Flysystem\UnableToMoveFile;
use League\Flysystem\UnableToWriteFile;

/**
 * A Flysystem adapter implementing caching with Drupal's Cache API.
 */
class DrupalCacheAdapter implements FilesystemAdapter {

  /**
   * The Flysystem adapter to cache data for.
   *
   * @var \League\Flysystem\FilesystemAdapter
   */
  protected $adapter;

  /**
   * The cache backend to store data in.
   *
   * @var \Drupal\flysystem\Flysystem\Adapter\CacheItemBackend
   */
  protected $cacheItemBackend;

  /**
   * The scheme of the stream wrapper used for this adapter.
   *
   * @var string
   */
  protected $scheme;

  /**
   * Constructs a new caching Flysystem adapter.
   *
   * @param string $scheme
   *   The scheme of the stream wrapper used for this adapter.
   * @param \League\Flysystem\FilesystemAdapter $adapter
   *   The flysystem adapter to cache data for.
   * @param \Drupal\flysystem\Flysystem\Adapter\CacheItemBackend $cacheItemBackend
   *   The cache backend to store data in.
   *
   * @todo Evaluate whether the FilesystemAdapter should actually be the Filesystem instead.
   * @see https://github.com/thephpleague/flysystem/blob/3.x/src/Filesystem.php
   */
  public function __construct($scheme, FilesystemAdapter $adapter, CacheItemBackend $cacheItemBackend) {
    $this->scheme = $scheme;
    $this->adapter = $adapter;
    $this->cacheItemBackend = $cacheItemBackend;
  }

  /**
   * {@inheritdoc}
   */
  public function write(string $path, string $contents, Config $config): void {
    try {
      $this->adapter->write($path, $contents, $config);
      $this->updateMetadata($path);
    }
    catch (UnableToWriteFile $exception) {
      throw $exception;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function writeStream($path, $resource, Config $config): void {
    try {
      $this->adapter->writeStream($path, $resource, $config);
      $this->updateMetadata($path);
    }
    catch (UnableToWriteFile $exception) {
      throw $exception;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function move($path, $newPath, $config): void {
    try {
      $this->adapter->move($path, $newPath, $config);
      $item = $this->cacheItemBackend->load($path);
      $newItem = clone $item;
      $this->cacheItemBackend->set($newPath, $newItem);
      $this->cacheItemBackend->delete($path);
    }
    catch (UnableToMoveFile $exception) {
      throw $exception;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function copy($path, $newPath, $config): void {
    // @todo See ::write for additional refactoring notes.
    try {
      $this->adapter->copy($path, $newPath, $config);
      $item = $this->cacheItemBackend->load($path);
      $newItem = clone $item;
      $this->cacheItemBackend->set($newPath, $newItem);
    }
    catch (UnableToCopyFile $exception) {
      throw $exception;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function delete($path): void {
    try {
      $this->adapter->delete($path);
      $this->cacheItemBackend->delete($path);
    }
    catch (UnableToDeleteFile $exception) {
      throw $exception;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function deleteDirectory($dirname): void {
    try {
      // Before the delete we need to know what files are in the directory.
      // Necessary to be able to delete them from the cache backend.
      $contents = $this->adapter->listContents($dirname, TRUE);

      $this->adapter->deleteDirectory($dirname);
      $directoryContents = iterator_to_array($contents);
      $paths = array_column($directoryContents, 'path');
      $this->cacheItemBackend->deleteMultiple($paths);
    }
    catch (UnableToDeleteDirectory $exception) {
      throw $exception;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function createDirectory($dirname, Config $config): void {
    try {
      $this->adapter->createDirectory($dirname, $config);
      $this->updateMetadata($dirname);
    }
    catch (UnableToCreateDirectory $exception) {
      throw $exception;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function setVisibility($path, $visibility): void {
    try {
      $this->adapter->setVisibility($path, $visibility);
      $this->updateMetadata($path);
    }
    catch (InvalidVisibilityProvided $exception) {
      throw $exception;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function fileExists($path): bool {
    try {
      return $this->adapter->fileExists($path);
    }
    catch (UnableToCheckExistence $exception) {
      throw $exception;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function directoryExists($path): bool {
    try {
      return $this->adapter->directoryExists($path);
    }
    catch (UnableToCheckExistence $exception) {
      throw $exception;
    }
  }

  /**
   * This has been deprecated at the Adapter level for Flysystem v3.
   *
   * We may not need this at all, or we may want to refactor to implement
   * the Filesystem instead of the adapter at this level, since that will
   * make Drupal Caching more universal across all adapters.
   *
   * @todo Reconsider implementation and possibly refactor to use Filesystem.
   */
  public function has($path): bool {
    // @todo This may not be needed, here, as ::has() no longer exists on the
    // adapter, it exists on the (League/Flysystem) Filesystem object.
    // Another case for using the Filesystem object instead of the adapter.
    if ($this->cacheItemBackend->has($path)) {
      return TRUE;
    }
    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function read(string $path): string {
    try {
      return $this->adapter->read($path);
    }
    catch (UnableToReadFile $exception) {
      throw $exception;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function readStream($path) {
    try {
      return $this->adapter->readStream($path);
    }
    catch (UnableToReadFile $exception) {
      throw $exception;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function listContents($directory = '', $recursive = FALSE): iterable {
    // Don't cache directory listings to avoid having to keep track of
    // incomplete cache entries.
    // @todo This could be a good place for a micro-cache?
    try {
      return $this->adapter->listContents($directory, $recursive);
    }
    catch (UnableToListContents $exception) {
      throw $exception;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function fileSize($path): FileAttributes {
    try {
      return $this->adapter->fileSize($path);
    }
    catch (UnableToRetrieveMetadata $exception) {
      throw $exception;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function mimeType($path): FileAttributes {
    try {
      return $this->adapter->mimeType($path);
    }
    catch (UnableToRetrieveMetadata $exception) {
      throw $exception;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function lastModified($path): FileAttributes {
    try {
      return $this->adapter->lastModified($path);
    }
    catch (UnableToRetrieveMetadata $exception) {
      throw $exception;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function visibility($path): FileAttributes {
    try {
      return $this->adapter->visibility($path);
    }
    catch (UnableToRetrieveMetadata $exception) {
      throw $exception;
    }
  }

  /**
   * Updates the metadata for a given path.
   *
   * @param string $path
   *   The path of file or directory.
   */
  protected function updateMetadata($path): void {
    // Note:  The following changes are based on Flysystem 3.x version of
    // the local adapter, this will need to be reviewed for relevancy for
    // all adapters to determine if this is the best overall approach.
    // @todo review code from here to end for relevance with all adapters.
    $entity = new FileAttributes($path);

    $metadata = [
      // FileAttributes object.
      'type' => $entity->type(),
      'path' => $path,
      'timestamp' => $entity->lastModified(),
    ];

    if ($metadata['type'] == 'file') {
      $metadata['size'] = $entity->fileSize();
    }

    if (!empty($metadata)) {
      $item = $this->cacheItemBackend->load($path);
      $item->updateMetadata($metadata);
      $this->cacheItemBackend->set($path, $item);
    }
  }

}
