<?php

declare(strict_types=1);

namespace Drupal\flysystem\Form;

use Drupal\Component\Plugin\Exception\PluginException;
use Drupal\Component\Render\MarkupInterface;
use Drupal\Component\Utility\Html;
use Drupal\Core\Entity\EntityForm;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Plugin\PluginFormInterface;
use Drupal\Core\Render\Markup;
use Drupal\flysystem\Entity\FlysystemAdapterConfig;
use Drupal\flysystem\FlysystemAdapterConfigInterface;
use Drupal\flysystem\FlysystemAdapterPluginManager;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Flysystem_adapter_config form.
 */
class FlysystemAdapterConfigForm extends EntityForm {
  /**
   * The adapter plugin manager.
   *
   * @var \Drupal\flysystem\FlysystemAdapterPluginManager
   */
  protected $subAdapterPluginManager;

  /**
   * The messenger.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * Constructs a FlysystemAdapterConfigForm object.
   *
   * @param \Drupal\flysystem\FlysystemAdapterPluginManager $subadapter_plugin_manager
   *   The adapter plugin manager.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger.
   */
  public function __construct(FlysystemAdapterPluginManager $subadapter_plugin_manager, MessengerInterface $messenger) {
    $this->subAdapterPluginManager = $subadapter_plugin_manager;
    $this->messenger = $messenger;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    $subadapter_plugin_manager = $container->get('plugin.manager.flysystem_adapter');
    $messenger = $container->get('messenger');

    return new static($subadapter_plugin_manager, $messenger);
  }

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state): array {
    // If the form is being rebuilt, rebuild the entity with the current form
    // values.
    if ($form_state->isRebuilding()) {
      $this->entity = $this->buildEntity($form, $form_state);
    }

    $form = parent::form($form, $form_state);

    /** @var \Drupal\flysystem\FlysystemAdapterConfigInterface $flysystem_adapter_config */
    $flysystem_adapter_config = $this->getEntity();

    // var_dump($flysystem_adapter_config); die;
    // Set the page title according to whether we are creating or editing the
    // adapter.
    if (!is_null($flysystem_adapter_config)) {
      if ($flysystem_adapter_config->isNew()) {
        $form['#title'] = $this->t('Configure a new Flysystem Adapter');
      }
      else {
        $form['#title'] = $this->t('Edit Flysystem Adapter %label', ['%label' => $flysystem_adapter_config->label()]);
      }
      $this->buildEntityConfigForm($form, $form_state, $flysystem_adapter_config);
      // Skip adding the adapter config form if we cleared the adapter form due
      // to an error.
      if ($form) {
        $this->buildSubAdapterConfigForm($form, $form_state, $flysystem_adapter_config);
      }

    }
    return $form;
  }

  /**
   * Builds the form for the basic adapter configuration properties.
   *
   * @param array $form
   *   The current form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   * @param \Drupal\flysystem\FlysystemAdapterConfigInterface $flysystem_adapter_config
   *   The server that is being created or edited.
   */
  protected function buildEntityConfigForm(array &$form, FormStateInterface $form_state, FlysystemAdapterConfigInterface $flysystem_adapter_config) {
    $form['label'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Scheme'),
      '#description' => $this->t('The scheme used to identify files managed with this adapter.'),
      '#maxlength' => 255,
      '#default_value' => $flysystem_adapter_config->label(),
      '#required' => TRUE,
    ];

    $form['id'] = [
      '#type' => 'machine_name',
      '#default_value' => $flysystem_adapter_config->id(),
      '#machine_name' => [
        'exists' => [FlysystemAdapterConfig::class, 'load'],
      ],
      '#disabled' => !$flysystem_adapter_config->isNew(),
    ];

    $form['status'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Enabled'),
      '#default_value' => $flysystem_adapter_config->status(),
    ];

    $form['description'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Description'),
      '#default_value' => $flysystem_adapter_config->getDescription(),
    ];

    $sub_adapters = $this->subAdapterPluginManager->getDefinitions();
    $options = [];
    $descriptions = [];
    foreach ($sub_adapters as $subadapter_id => $definition) {
      $config = $subadapter_id === $flysystem_adapter_config->getSubAdapterId() ? $flysystem_adapter_config->getSubAdapterConfig() : [];
      $config['#flysystem_adapter_config'] = $flysystem_adapter_config;
      try {
        /** @var \Drupal\flysystem\FlysystemSubAdapterInterface $sub_adapter */
        $sub_adapter = $this->subAdapterPluginManager
          ->createInstance($subadapter_id, $config);
      }
      catch (PluginException) {
        // @todo add graceful error handling.
        continue;
      }
      $options[$subadapter_id] = $this->escapeHtml($sub_adapter->label());
      $descriptions[$subadapter_id]['#description'] = $this->escapeHtml($sub_adapter->getDescription());

    }
    asort($options, SORT_NATURAL | SORT_FLAG_CASE);
    if ($options) {
      if (count($options) == 1) {
        $flysystem_adapter_config->set('sub_adapter', key($options));
      }
      $form['sub_adapter'] = [
        '#type' => 'radios',
        '#title' => $this->t('Flysystem Adapters'),
        '#description' => $this->t('Choose a Flysystem Adapter type to configure.'),
        '#options' => $options,
        '#default_value' => $flysystem_adapter_config->getSubAdapterId(),
        '#required' => TRUE,
        '#disabled' => !$flysystem_adapter_config->isNew(),
        '#ajax' => [
          'callback' => [get_class($this), 'buildAjaxBackendConfigForm'],
          'wrapper' => 'flysystem-adapter-config-form',
          'method' => 'replace',
          'effect' => 'fade',
        ],
      ];
      $form['sub_adapter'] += $descriptions;
    }
    /*
    else {
    // @todo update with link to documentation on D.O.
    $url = 'https://www.drupal.org/docs/8/modules/search-api/getting-started/server-backends-and-features';
    $args[':url'] = Url::fromUri($url)->toString();
    $error = $this->t('There are no Flysystem Adapters available to configure. \
    Install a <a href=":url">module that provides a Flysystem Adapter</a> \
    to proceed.', $args);
    $this->messenger->addError($error);
    $form = [];
    }
     */
  }

  /**
   * {@inheritdoc}
   */
  protected function actions(array $form, FormStateInterface $form_state) {
    if ($form === []) {
      return [];
    }

    return parent::actions($form, $form_state);
  }

  /**
   * Builds the adapter-specific configuration form.
   *
   * @param array $form
   *   The current form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   * @param \Drupal\flysystem\FlysystemAdapterConfigInterface $flysystem_adapter_config
   *   The adapter that is being created or edited.
   */
  public function buildSubAdapterConfigForm(array &$form, FormStateInterface $form_state, FlysystemAdapterConfigInterface $flysystem_adapter_config) {
    $form['sub_adapter_config'] = [];

    if ($flysystem_adapter_config->hasValidSubAdapter()) {
      $sub_adapter_plugin = $flysystem_adapter_config->getSubAdapter();
      $form_state->set('sub_adapter', $sub_adapter_plugin->getPluginId());
      $form_state->set(['sub_adapter_config', 'schema'], $flysystem_adapter_config->getSubAdapterConfigItem('schema'));

      if ($sub_adapter_plugin instanceof PluginFormInterface) {
        if ($form_state->isRebuilding()) {
          $this->messenger->addWarning($this->t('Configure the selected Flysystem Adapter.'));
        }

        // Attach the subadapter plugin configuration form.
        $form['sub_adapter_config'] = $sub_adapter_plugin->buildSubConfigurationForm($form['sub_adapter_config'], $form_state, $flysystem_adapter_config);

        // Modify the backend plugin configuration container element.
        $form['sub_adapter_config']['#type'] = 'details';
        $form['sub_adapter_config']['#title'] = $this->t('Configure %plugin configuration', ['%plugin' => $sub_adapter_plugin->label()]);
        $form['sub_adapter_config']['#open'] = TRUE;
        // var_dump($form['sub_adapter_config']); die;.
      }
    }
    // Only notify the user of a missing backend plugin if we're editing an
    // existing server.
    elseif (!$flysystem_adapter_config->isNew()) {
      $this->messenger->addError($this->t('The Flysystem adapter plugin is missing or invalid.'));
      return;
    }
    $form['sub_adapter_config'] += [
      '#type' => 'container',
    ];
    $form['sub_adapter_config']['#attributes']['id'] = 'flysystem-adapter-config-form';
    $form['sub_adapter_config']['#tree'] = TRUE;
  }

  /**
   * Handles switching the selected adapter plugin.
   *
   * @param array $form
   *   The current form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   *
   * @return array
   *   The part of the form to return as AJAX.
   */
  public static function buildAjaxAdapterConfigForm(array $form, FormStateInterface $form_state) {
    // The work is already done in form(), where we rebuild the entity according
    // to the current form values and then create the backend configuration form
    // based on that. So we just need to return the relevant part of the form
    // here.
    return $form['sub_adapter_config'];
  }

  /**
   * {@inheritdoc}
   */
  public function save(array $form, FormStateInterface $form_state): int {
    $result = parent::save($form, $form_state);
    $this->messenger->addStatus($this->t('The Flysystem adapter was successfully saved.'));
    $message_args = ['%label' => $this->entity->label()];
    $message = '';
    switch ($result) {
      case \SAVED_NEW:
        $message = $this->t('Created new example %label.', $message_args);
        break;

      case \SAVED_UPDATED:
        $message = $this->t('Updated example %label.', $message_args);
        break;
    }
    $this->messenger()->addStatus($message);
    $form_state->setRedirectUrl($this->entity->toUrl('collection'));
    return $result;
  }

  /**
   * Escapes HTML special characters in plain text, if necessary.
   *
   * @param string|\Drupal\Component\Render\MarkupInterface $text
   *   The text to escape.
   *
   * @return \Drupal\Component\Render\MarkupInterface
   *   If a markup object was passed as $text, it is returned as-is. Otherwise,
   *   the text is escaped and returned
   */
  private function escapeHtml($text) {
    if ($text instanceof MarkupInterface) {
      return $text;
    }

    return Markup::create(Html::escape((string) $text));
  }

}
