<?php

namespace Drupal\Tests\flysystem\Unit\Flysystem\Adapter;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Cache\MemoryBackend;
use Drupal\flysystem\Flysystem\Adapter\CacheItemBackend;
use Drupal\flysystem\Flysystem\Adapter\DrupalCacheAdapter;
use Drupal\Tests\UnitTestCase;
use League\Flysystem\Config;
use League\Flysystem\FileAttributes;
use League\Flysystem\StorageAttributes;
use League\Flysystem\Local\LocalFilesystemAdapter;
use League\Flysystem\Visibility;
use League\Flysystem\UnableToCreateDirectory;
use League\Flysystem\UnableToMoveFile;
use League\Flysystem\UnableToReadFile;
use League\Flysystem\UnableToRetrieveMetadata;
use League\Flysystem\UnableToSetVisibility;
use League\Flysystem\UnableToWriteFile;

/**
 * Test the Drupal Cache Adapter.
 *
 * @group flysystem
 *
 * @coversDefaultClass \Drupal\flysystem\Flysystem\Adapter\DrupalCacheAdapter
 * @covers \Drupal\flysystem\Flysystem\Adapter\DrupalCacheAdapter
 */
class DrupalCacheAdapterTest extends UnitTestCase {

  /**
   * URI scheme to use for testing.
   *
   * @var string
   */
  const SCHEME = 'test_scheme';

  /**
   * The main test file.
   *
   * @var string
   */
  const FILE = 'test.txt';

  /**
   * The directory for the LocalAdapter configuration.
   *
   * @var string
   */
  const ROOT = 'sites/default/files/flysystem/testing';

  /**
   * The directory for the LocalAdapter that fails tests.
   */
  const FAKEROOT = '/';

  /**
   * The wrapped Flysystem adapter.
   *
   * @var \League\Flysystem\FilesystemAdapter|\PHPUnit_Framework_MockObject_MockObject
   */
  protected $adapter;

  /**
   * The cache adapter under test.
   *
   * @var \Drupal\flysystem\Flysystem\Adapter\DrupalCacheAdapter
   */
  protected $cacheAdapter;

  /**
   * The flysystem backend for testing.
   *
   * @var \Drupal\flysystem\Flysystem\Adapter\CacheItemBackend
   */
  protected $cacheItemBackend;

  /**
   * Timestamp of test run.
   *
   * @var int
   */
  protected $time;

  /**
   * {@inheritdoc}
   */
  public function setUp(): void {
    parent::setUp();
    $time = $this->prophesize(TimeInterface::class)->reveal();
    $this->cacheItemBackend = new CacheItemBackend(static::SCHEME, new MemoryBackend($time));
    $this->adapter = new LocalFilesystemAdapter(
      static::ROOT,
    );
    $this->cacheAdapter = new DrupalCacheAdapter(static::SCHEME, $this->adapter, $this->cacheItemBackend);
  }

  /**
   * {@inheritdoc}
   */
  public function tearDown(): void {
    parent::tearDown();
    $this->delTree(static::ROOT);
  }

  /**
   * @covers ::write
   */
  public function testWrite() {
    $config = new Config(['visibility' => 'public']);
    // Test Cache adapter write does not fail.
    $this->cacheAdapter->write(static::FILE, 'contents', $config);
    $this->assertTrue($this->cacheAdapter->fileExists(static::FILE));
    $fileAttributes = $this->cacheAdapter->visibility(static::FILE);
    $this->assertSame(Visibility::PUBLIC, $fileAttributes->visibility());
    // Test Cache Adapter Write fails.
    $this->expectException(UnableToWriteFile::class);
    $this->adapter = new LocalFilesystemAdapter(static::FAKEROOT);
    $this->cacheAdapter = new DrupalCacheAdapter(static::SCHEME, $this->adapter, $this->cacheItemBackend);
    $this->cacheAdapter->write(static::FILE, 'contents', $config);
    $this->assertFalse($this->cacheAdapter->fileExists(static::FILE));
  }

  /**
   * Tests creating a public file stream.
   *
   * @covers ::writeStream
   * @covers ::lastModified
   */
  public function testWriteStream() {
    $config = new Config([]);
    $stream = fopen('data:text/plain,contents', 'rb');
    $this->cacheAdapter->writeStream(static::FILE, $stream, $config);
    $this->time = time();
    // Test lastModified on.
    $fileAttributes = $this->cacheAdapter->lastModified(static::FILE);
    $this->assertSame($this->time, $fileAttributes->lastModified());
    // Failing to write a file with a stream.
    $this->expectException(UnableToWriteFile::class);
    $this->adapter = new LocalFilesystemAdapter(static::FAKEROOT);
    $this->cacheAdapter = new DrupalCacheAdapter(static::SCHEME, $this->adapter, $this->cacheItemBackend);
    try {
      // stream_with_contents()
      $stream = fopen('php://temp', 'w+b');
      fwrite($stream, 'something');
      rewind($stream);
      $this->cacheAdapter->writeStream(static::FILE, $stream, new Config());
    } finally {
      isset($stream) && is_resource($stream) && fclose($stream);
    }
  }

  /**
   * Tests moving a file.
   *
   * @covers ::write
   * @covers ::visibility
   * @covers ::move
   */
  public function testMove() {
    $config = new Config([]);
    $this->cacheAdapter->write(static::FILE, 'contents', $config);
    $fileAttributes = $this->cacheAdapter->visibility(static::FILE);
    $this->assertSame(Visibility::PUBLIC, $fileAttributes->visibility());
    // Test Cache adapter move does not fail.
    $this->cacheAdapter->move(static::FILE, 'new.txt', $config);
    // Test that old file information is no longer cached.
    $this->assertFalse($this->cacheItemBackend->has(static::FILE));
    // Test that new file information is now cached.
    $this->assertTrue($this->cacheItemBackend->has('new.txt'));
    // Check visibility of new file.
    $fileAttributes = $this->cacheAdapter->visibility('new.txt');
    $this->assertSame(Visibility::PUBLIC, $fileAttributes->visibility());

    // Create new files to move.
    $this->cacheAdapter->write('first.txt', 'contents', new Config(['visibility' => 'public']));
    // Check that new file exists.
    $this->assertFileExists(static::ROOT . '/first.txt');
    // Check that new file has appropriate permissions.
    $this->assertFileHasPermissions(static::ROOT . '/first.txt', 0644);
    // Move file, make sure it is private.
    $this->cacheAdapter->move('first.txt', 'second.txt', new Config(['visibility' => 'private']));
    // Check that new file exists.
    $this->assertFileExists(static::ROOT . '/second.txt');
    // Check that permissions on new file is private.
    $this->assertFileHasPermissions(static::ROOT . '/second.txt', 0600);
    // Test move failed.
    $this->expectException(UnableToMoveFile::class);
    $this->adapter = new LocalFilesystemAdapter(static::FAKEROOT);
    $this->cacheAdapter = new DrupalCacheAdapter(static::SCHEME, $this->adapter, $this->cacheItemBackend);
    $this->cacheAdapter->move(static::FILE, 'new.txt', $config);

  }

  /**
   * Tests copying a file.
   *
   * @covers ::write
   * @covers ::copy
   * @covers ::visibility
   */
  public function testCopy() {
    $config = new Config([]);
    // Create initial file.
    $this->cacheAdapter->write(static::FILE, 'contents', $config);
    // Check that new file is publicly visible.
    $fileAttributes = $this->cacheAdapter->visibility(static::FILE);
    $this->assertSame(Visibility::PUBLIC, $fileAttributes->visibility());
    // Test Cache adapter copy does not fail.
    $this->cacheAdapter->copy(static::FILE, 'new.txt', $config);
    // Check that both files are cached.
    $this->assertTrue($this->cacheItemBackend->has(static::FILE));
    $this->assertTrue($this->cacheItemBackend->has('new.txt'));
    // Check visibility on new file.
    $fileAttributes = $this->cacheAdapter->visibility('new.txt');
    $this->assertSame(Visibility::PUBLIC, $fileAttributes->visibility());
    // Copy failed test.
    $this->expectException(UnableToMoveFile::class);
    $this->adapter = new LocalFilesystemAdapter(static::FAKEROOT);
    $this->cacheAdapter = new DrupalCacheAdapter(static::SCHEME, $this->adapter, $this->cacheItemBackend);
    $this->cacheAdapter->move(static::FILE, 'new.txt', $config);

  }

  /**
   * Tests deleting a file.
   *
   * @covers ::delete
   */
  public function testDelete() {
    $config = new Config([]);
    $this->cacheAdapter->write(static::FILE, 'contents', $config);
    $fileAttributes = $this->cacheAdapter->visibility(static::FILE);
    $this->assertSame(Visibility::PUBLIC, $fileAttributes->visibility());

    $this->cacheAdapter->delete(static::FILE);
    $this->assertFalse($this->cacheItemBackend->has(static::FILE));
  }

  /**
   * Tests deleting a file directory.
   *
   * @covers ::deleteDirectory
   */
  public function testDeleteDirectory() {
    mkdir(static::ROOT . '/directory/subdir/', 0744, TRUE);
    $this->assertDirectoryExists(static::ROOT . '/directory/subdir/');
    file_put_contents(static::ROOT . '/directory/subdir/file.txt', 'content');
    symlink(static::ROOT . '/directory/subdir/file.txt', static::ROOT . '/directory/subdir/link.txt');
    $this->cacheAdapter->deleteDirectory('directory/subdir');
    $this->assertDirectoryDoesNotExist(static::ROOT . '/directory/subdir/');
    $this->cacheAdapter->deleteDirectory('directory');
    $this->assertDirectoryDoesNotExist(static::ROOT . '/directory/');

  }

  /**
   * Tests deleting a file directory.
   *
   * @covers ::createDirectory
   */
  public function testCreateDirectory() {
    // Create public directory.
    $this->cacheAdapter->createDirectory('public', new Config(['visibility' => 'public']));
    $this->assertDirectoryExists(static::ROOT . '/public');
    $this->assertFileHasPermissions(static::ROOT . '/public', 0755);
    // Create Private directory.
    $this->cacheAdapter->createDirectory('private', new Config(['visibility' => 'private']));
    $this->assertDirectoryExists(static::ROOT . '/private');
    $this->assertFileHasPermissions(static::ROOT . '/private', 0700);
    // Unable to Create Directory.
    $this->expectException(UnableToCreateDirectory::class);
    $this->adapter = new LocalFilesystemAdapter(static::FAKEROOT);
    $this->cacheAdapter = new DrupalCacheAdapter(static::SCHEME, $this->adapter, $this->cacheItemBackend);
    $this->cacheAdapter->createDirectory('/something/', new Config());

  }

  /**
   * Tests visibility checking.
   *
   * @covers ::visibility
   */
  public function testVisibility() {
    $this->cacheAdapter->write('public.txt', 'contents', new Config(['visibility' => 'public']));
    $this->assertTrue($this->cacheItemBackend->has('public.txt'));
    $this->assertEquals('public', $this->cacheAdapter->visibility('public.txt')->visibility());
    $this->cacheAdapter->write('private.txt', 'contents', new Config(['visibility' => 'private']));
    $this->assertTrue($this->cacheItemBackend->has('private.txt'));
    $this->assertEquals('private', $this->cacheAdapter->visibility('private.txt')->visibility());
    // Unable to retrieve visibility.
    $this->expectException(UnableToRetrieveMetadata::class);
    $this->assertFalse($this->cacheItemBackend->has('something.txt'));
    $this->cacheAdapter->visibility('something.txt');
  }

  /**
   * @covers ::setVisibility
   */
  public function testSetVisibility() {
    // Create a file to test against.
    $this->cacheAdapter->write(static::FILE, 'contents', new Config(['visibility' => 'public']));
    $this->assertEquals(Visibility::PUBLIC, $this->cacheAdapter->visibility(static::FILE)->visibility());
    // Set visibility to private and test.
    $this->cacheAdapter->setVisibility(static::FILE, Visibility::PRIVATE);
    $this->assertEquals(Visibility::PRIVATE, $this->cacheAdapter->visibility(static::FILE)->visibility());
    // Set visibility to public and test.
    $this->cacheAdapter->setVisibility(static::FILE, Visibility::PUBLIC);
    $this->assertEquals(Visibility::PUBLIC, $this->cacheAdapter->visibility(static::FILE)->visibility());
    // Fail setting visibility.
    $this->expectException(UnableToSetVisibility::class);
    $this->cacheAdapter->setVisibility('this-path-does-not-exists.txt', Visibility::PRIVATE);
  }

  /**
   * @covers ::directoryExists
   */
  public function testDirectoryExists() {
    // Test that directory does not exist.
    $this->assertFalse($this->cacheAdapter->directoryExists(static::ROOT));
    // Create directory and test that it exists.
    $this->cacheAdapter->createDirectory(static::ROOT, new Config([]));
    $this->assertTrue($this->cacheAdapter->directoryExists(static::ROOT));
    // Create a file.
    $this->cacheAdapter->write('test.txt', 'content', new Config([]));
    // Test that it is a file and not a directory.
    $this->assertTrue($this->cacheAdapter->fileExists('test.txt'));
    $this->assertFalse($this->cacheAdapter->directoryExists('test.txt'));
  }

  /**
   * @covers ::fileExists
   */
  public function testFileExists() {

    // Does not exist before creation.
    self::assertFalse($this->cacheAdapter->fileExists(static::FILE));
    $this->cacheAdapter->write(static::FILE, 'content', new Config([]));
    // A file exists after creation.
    self::assertTrue($this->cacheAdapter->fileExists(static::FILE));

    // A file no longer exists after deletion.
    $this->cacheAdapter->delete(static::FILE);
    self::assertFalse($this->cacheAdapter->fileExists(static::FILE));
  }

  /**
   * Tests file loading from cache success.
   */
  public function testHas() {
    $cache_item = $this->cacheItemBackend->load(static::FILE);
    $this->cacheItemBackend->set(static::FILE, $cache_item);
    $this->assertTrue($this->cacheItemBackend->has(static::FILE));
    // Tests cache does not have an item.
    $this->cacheItemBackend->delete(static::FILE);
    $this->assertFalse($this->cacheItemBackend->has(static::FILE));
  }

  /**
   * Tests reading a file.
   *
   * @covers ::visibility
   * @covers ::read
   */
  public function testRead() {

    $config = new Config([]);
    $this->cacheAdapter->write(static::FILE, 'contents', $config);

    $fileAttributes = $this->cacheAdapter->visibility(static::FILE);
    $this->assertSame(Visibility::PUBLIC, $fileAttributes->visibility());
    $this->assertSame($this->cacheAdapter->read(static::FILE), 'contents');
    // Fail to read.
    $this->expectException(UnableToReadFile::class);
    $this->adapter = new LocalFilesystemAdapter(static::FAKEROOT);
    $this->cacheAdapter = new DrupalCacheAdapter(static::SCHEME, $this->adapter, $this->cacheItemBackend);
    $this->cacheAdapter->read(static::FILE);
  }

  /**
   * Tests reading a stream.
   *
   * @covers ::write
   * @covers ::readStream
   */
  public function testReadStream() {
    $config = new Config([]);

    $this->cacheAdapter->write(static::FILE, 'contents', $config);
    $contents = $this->cacheAdapter->readStream(static::FILE);
    $this->assertIsResource($contents);
    $fileContents = stream_get_contents($contents);
    fclose($contents);
    $this->assertEquals('contents', $fileContents);
    // Unable to read stream.
    $this->expectException(UnableToReadFile::class);
    $this->adapter = new LocalFilesystemAdapter(static::FAKEROOT);
    $this->cacheAdapter = new DrupalCacheAdapter(static::SCHEME, $this->adapter, $this->cacheItemBackend);
    $this->cacheAdapter->readStream(static::FILE);
  }

  /**
   * Tests listing contents of a directory.
   *
   * @covers ::listContents
   */
  public function testListContents() {
    $config = new Config([]);
    $this->cacheAdapter->write('directory/filename.txt', 'content', $config);
    $this->cacheAdapter->write('filename.txt', 'content', $config);
    /** @var Traversable $contentListing */
    $contentListing = $this->cacheAdapter->listContents('/', FALSE);
    $contents = iterator_to_array($contentListing);
    $this->assertCount(2, $contents);
    $this->assertContainsOnlyInstancesOf(StorageAttributes::class, $contents);
    // List contents of a non existent directory.
    $this->adapter = new LocalFilesystemAdapter(static::FAKEROOT);
    $this->cacheAdapter = new DrupalCacheAdapter(static::SCHEME, $this->adapter, $this->cacheItemBackend);
    /** @var Traversable $contentListing */
    $contentListing = $this->cacheAdapter->listContents('/directory/', FALSE);
    $contents = iterator_to_array($contentListing);
    $this->assertCount(0, $contents);
  }

  /**
   * Tests listing contents of a directory recursively.
   *
   * @covers ::listContents
   */
  public function testListContentsRecursively() {
    $config = new Config([]);
    $this->cacheAdapter->write('directory/filename.txt', 'content', $config);
    $this->cacheAdapter->write('filename.txt', 'content', $config);
    /** @var Traversable $contentListing */
    $contentListing = $this->cacheAdapter->listContents('/', TRUE);
    $contents = iterator_to_array($contentListing);
    $this->assertCount(3, $contents);
    $this->assertContainsOnlyInstancesOf(StorageAttributes::class, $contents);
  }

  /**
   * @covers ::fileSize
   */
  public function testFileSize() {
    $config = new Config([]);
    $this->cacheAdapter->write(static::FILE, 'contents', $config);
    // Test for file Size.
    $attributes = $this->cacheAdapter->fileSize(static::FILE);
    $this->assertInstanceOf(FileAttributes::class, $attributes);
    $this->assertEquals(8, $attributes->fileSize());
    // Failed test for directory size.
    $this->expectException(UnableToRetrieveMetadata::class);
    $this->cacheAdapter->createDirectory(static::FAKEROOT, $config);
    $this->cacheAdapter->fileSize(static::FILE . '/');
    // Fail retrieving size on non-existent file.
    $this->expectException(UnableToRetrieveMetadata::class);
    $this->cacheAdapter->fileSize('non-existing-file.txt');

  }

  /**
   * @covers ::mimeType
   */
  public function testMimeType() {
    // Test for exact mimetype match.
    $config = new Config([]);
    $this->cacheAdapter->write('something.csv', 'contents', $config);
    $mimeType = $this->cacheAdapter->mimeType('something.csv');
    $this->assertEquals('text/csv', $mimeType->mimeType());
    // Test for partial match on mimetype...svg.
    $fileContents = (string) file_get_contents(__DIR__ . '/../../../../fixtures/files/flysystem.svg');
    $this->cacheAdapter->write('flysystem.svg', $fileContents, $config);
    $this->assertStringStartsWith('image/svg+xml', $this->cacheAdapter->mimeType('flysystem.svg')->mimeType());
    // Test for non-existent file.
    $this->expectException(UnableToRetrieveMetadata::class);
    $this->cacheAdapter->mimeType('non-existing-file.txt');
    // Test for unknown file type.
    $fileContents = (string) file_get_contents(__DIR__ . '/../../../../fixtures/files/unknown-mime-type.md5');
    $this->cacheAdapter->write('unknown-mime-type.md', $fileContents, $config);
    $this->expectException(UnableToRetrieveMetadata::class);
    $this->cacheAdapter->mimeType('unknown-mime-type.md5');
  }

  /**
   * @covers ::lastModified
   */
  public function testLastModified() {
    $config = new Config([]);
    $this->cacheAdapter->write(static::FILE, 'contents', $config);
    $attributes = $this->cacheAdapter->lastModified(static::FILE);
    $this->assertInstanceOf(FileAttributes::class, $attributes);
    $this->assertIsInt($attributes->lastModified());
    $this->assertTrue($attributes->lastModified() > time() - 30);
    $this->assertTrue($attributes->lastModified() < time() + 30);
    // Test for non-existent file.
    $this->expectException(UnableToRetrieveMetadata::class);
    $this->cacheAdapter->lastModified('non-existing-file.txt');
  }

  /**
   * Recursively deletes a directory.
   *
   * @param string $path
   *   Directory path to delete.
   *
   * @return bool
   *   True on success, False on failure.
   */
  protected function delTree($path) {
    $files = array_diff(scandir($path), ['.', '..']);
    foreach ($files as $file) {
      (is_dir("$path/$file")) ? $this->delTree("$path/$file") : unlink("$path/$file");
    }
    return rmdir($path);
  }

  /**
   * Checks that a file has indicated permissions.
   *
   * @param string $file
   *   File to check.
   * @param int $expectedPermissions
   *   Permissions to check on file.
   */
  private function assertFileHasPermissions(string $file, int $expectedPermissions): void {
    clearstatcache(FALSE, $file);
    $permissions = fileperms($file) & 0777;
    $this->assertEquals($expectedPermissions, $permissions);
  }

}
