<?php

namespace Drupal\Tests\flysystem\Unit\Flysystem;

use Drupal\Core\DependencyInjection\ContainerBuilder;
use Drupal\Core\Logger\RfcLogLevel;
use Drupal\Core\Routing\UrlGeneratorInterface;
use Drupal\Core\Site\Settings;
use Drupal\flysystem\Flysystem\Adapter\MissingAdapter;
use Drupal\flysystem\Flysystem\Local;
use Drupal\Tests\UnitTestCase;
use League\Flysystem\Local\LocalFilesystemAdapter;
use Prophecy\Argument;

/**
 * @coversDefaultClass \Drupal\flysystem\Flysystem\Local
 * @group flysystem
 */
class LocalTest extends UnitTestCase {

  /**
   * {@inheritdoc}
   */
  public function setUp(): void {
    parent::setUp();
    $GLOBALS['base_url'] = 'http://example.com';
    $GLOBALS['FILES_DIRECTORY'] = 'sites/default/files/testing/';
    $container = new ContainerBuilder();
    $url_generator = $this->prophesize(UrlGeneratorInterface::class);
    $url_generator->generateFromRoute(Argument::cetera())->willReturn('download');
    $container->set('url_generator', $url_generator->reveal());
    \Drupal::setContainer($container);
    (new LocalFilesystemAdapter($GLOBALS['FILES_DIRECTORY'] . 'foo/bar'))->deleteDirectory('');
    @rmdir($GLOBALS['FILES_DIRECTORY'] . 'foo/bar');
    @rmdir($GLOBALS['FILES_DIRECTORY'] . 'foo');
    mkdir($GLOBALS['FILES_DIRECTORY'] . 'foo');
    mkdir($GLOBALS['FILES_DIRECTORY'] . 'foo/bar');
    touch($GLOBALS['FILES_DIRECTORY'] . 'test.txt');
  }

  /**
   * {@inheritdoc}
   */
  public function tearDown(): void {
    (new LocalFilesystemAdapter($GLOBALS['FILES_DIRECTORY'] . 'foo'))->deleteDirectory('');

    unlink($GLOBALS['FILES_DIRECTORY'] . 'test.txt');

    @unlink($GLOBALS['FILES_DIRECTORY'] . 'does_not_exist/.htaccess');
    @rmdir($GLOBALS['FILES_DIRECTORY'] . 'does_not_exist');
    $this->delTree($GLOBALS['FILES_DIRECTORY']);

  }

  /**
   * @covers ::__construct
   * @covers ::create
   */
  public function testCreateReturnsPlugin() {
    $container = new ContainerBuilder();
    $settings = new Settings([]);
    $container->set('settings', $settings);

    $configuration = ['root' => $GLOBALS['FILES_DIRECTORY'] . 'foo/bar'];

    $this->assertInstanceOf(Local::class, Local::create($container, $configuration, '', []));
  }

  /**
   * @covers ::getAdapter
   * @covers ::ensureDirectory
   */
  public function testReturnsLocalFilesystemAdapter() {
    $this->assertInstanceOf(LocalFilesystemAdapter::class, (new Local($GLOBALS['FILES_DIRECTORY'] . 'foo/bar', FALSE))->getAdapter());
  }

  /**
   * @covers ::getAdapter
   * @covers ::ensureDirectory
   */
  public function testMissingAdapterReturnedWhenPathIsFile() {

    $this->assertInstanceOf(MissingAdapter::class, (new Local($GLOBALS['FILES_DIRECTORY'] . 'test.txt'))->getAdapter());
  }

  /**
   * @covers ::getExternalUrl
   */
  public function testReturnsValidLocalUrl() {
    $plugin = new Local($GLOBALS['FILES_DIRECTORY'] . 'foo/bar', FALSE);
    $this->assertSame('download', $plugin->getExternalUrl('uri://test.html'));
  }

  /**
   * @covers ::getExternalUrl
   */
  public function testReturnsValidExternalUrl() {
    $plugin = new Local($GLOBALS['FILES_DIRECTORY'] . 'foo/bar', TRUE);
    $this->assertSame('http://example.com/' . $GLOBALS['FILES_DIRECTORY'] . 'foo/bar/test%20thing.html', $plugin->getExternalUrl('uri://test thing.html'));
  }

  /**
   * @covers ::ensure
   * @covers ::ensureDirectory
   */
  public function testDirectoryIsAutoCreatedAndHtaccessIsWritten() {
    new Local($GLOBALS['FILES_DIRECTORY'] . 'does_not_exist');
    $this->assertTrue(is_dir($GLOBALS['FILES_DIRECTORY'] . 'does_not_exist'));
    $this->assertTrue(is_file($GLOBALS['FILES_DIRECTORY'] . 'does_not_exist/.htaccess'));

  }

  /**
   * @covers ::ensure
   * @covers ::writeHtaccess
   */
  public function testHtaccessNotOverwritten() {
    file_put_contents($GLOBALS['FILES_DIRECTORY'] . 'foo/bar/.htaccess', 'htaccess content');

    $result = (new Local($GLOBALS['FILES_DIRECTORY'] . 'foo/bar'))->ensure();
    $this->assertCount(1, $result);
    $this->assertSame(RfcLogLevel::INFO, $result[0]['severity']);
    $this->assertSame('htaccess content', file_get_contents($GLOBALS['FILES_DIRECTORY'] . 'foo/bar/.htaccess'));
  }

  /**
   * @covers ::ensure
   * @covers ::writeHtaccess
   */
  public function testHtaccessNotOverwrittenAndFails() {
    @mkdir('foo');
    @mkdir('foo/bar');
    @mkdir('foo/bar/.htaccess', 0777, TRUE);
    $result = (new Local('foo/bar'))->ensure(TRUE);
    $this->assertCount(1, $result);
    $this->assertSame('https://www.drupal.org/SA-CORE-2013-003', $result[0]['context']['@url']);
  }

  /**
   * @covers ::ensure
   * @covers ::writeHtaccess
   */
  public function testEnsureReturnsErrorWhenCantCreateDir() {
    $result = (new Local($GLOBALS['FILES_DIRECTORY'] . 'test.txt'))->ensure();
    $this->assertSame($GLOBALS['FILES_DIRECTORY'] . 'test.txt', $result[0]['context']['%root']);
  }

  /**
   * Recursively deletes a directory.
   *
   * @param string $dir
   *   Directory to delete.
   */
  protected function delTree($dir) {
    $files = array_diff(scandir($dir), ['.', '..']);
    foreach ($files as $file) {
      (is_dir("$dir/$file")) ? $this->delTree("$dir/$file") : unlink("$dir/$file");
    }
    return rmdir($dir);
  }

}
